from aiozk import WatchEvent

from .base_watcher import BaseWatcher


class DataWatcher(BaseWatcher):

    watched_event = WatchEvent.DATA_CHANGED

    async def fetch(self, path):
        data = await self.client.get_data(path=path, watch=True)
        return data



import asyncio
import logging
import uuid
import pytest

from aiozk import WatchEvent


logging.getLogger('asyncio').setLevel(logging.DEBUG)


@pytest.fixture
async def full_zk(zk, path):
    child_1 = f'{path}/{uuid.uuid4().hex}'
    child_2 = f'{path}/{uuid.uuid4().hex}'
    await zk.create(path)
    await zk.create(child_1)
    await zk.create(child_2)
    yield zk
    await zk.delete(child_2)
    await zk.delete(child_1)
    await zk.delete(path)


@pytest.mark.asyncio
async def test_children(full_zk, path):
    resp = await full_zk.get_children(path)
    assert len(resp) == 2


@pytest.mark.asyncio
async def test_cancel_crash(zk, path):
    async def wait_loop():
        while 1:
            await zk.wait_for_event(WatchEvent.DATA_CHANGED, path)

    f = asyncio.ensure_future(wait_loop())
    f.cancel()



import asyncio
import collections
import inspect
import logging

from aiozk import exc

from .recipe import Recipe


log = logging.getLogger(__name__)


def maybe_future(fut, loop):
    if inspect.isawaitable(fut):
        asyncio.ensure_future(fut, loop=loop)


class BaseWatcher(Recipe):

    watched_event = None

    def __init__(self, *args, **kwargs):
        super(BaseWatcher, self).__init__(*args, **kwargs)
        self.callbacks = collections.defaultdict(set)
        self.loops = {}

    def add_callback(self, path, callback):
        self.callbacks[path].add(callback)

        if path not in self.loops:
            self.loops[path] = asyncio.ensure_future(self.watch_loop(path), loop=self.client.loop)

    def remove_callback(self, path, callback):
        self.callbacks[path].discard(callback)

        if not self.callbacks[path]:
            self.loops.pop(path).cancel()

    async def fetch(self, path):
        raise NotImplementedError

    def cancel(self):
        for loop in self.loops.values():
            loop.cancel()

    async def watch_loop(self, path):
        while self.callbacks[path]:
            log.debug("Fetching data for %s", path)
            try:
                result = await self.fetch(path)
            except exc.NoNode:
                for callback in self.callbacks[path]:
                    maybe_future(callback(exc.NoNode), loop=self.client.loop)
                return
            for callback in self.callbacks[path]:
                maybe_future(callback(result), loop=self.client.loop)
            await self.client.wait_for_event(self.watched_event, path)



import asyncio
import uuid

import pytest

from ..exc import NoNode


@pytest.fixture
async def data_watcher(zk, path):
    await zk.create(path)
    watcher = zk.recipes.DataWatcher()
    watcher.set_client(zk)
    yield watcher
    watcher.cancel()
    await zk.delete(path)


@pytest.mark.asyncio
async def test_data_watch(zk, path, data_watcher):
    data = []
    ready = asyncio.Event()
    test_data = b'test' * 1000

    async def data_callback(d):
        data.append(d)
        ready.set()

    data_watcher.add_callback(path, data_callback)
    assert data == []
    await zk.set_data(path, test_data)
    await asyncio.wait_for(ready.wait(), timeout=0.1)
    assert ready.is_set()
    assert data == [test_data]
    asyncio.Future()


@pytest.mark.asyncio
async def test_data_watch_no_node(zk, path, data_watcher):
    random_path = path + uuid.uuid4().hex
    is_finished = asyncio.Future()

    async def stub_callback(d):
        assert d == NoNode
        is_finished.set_result(True)

    data_watcher.add_callback(random_path, stub_callback)
    await asyncio.wait_for(is_finished, 0.1)


@pytest.fixture
def child1(path):
    yield f'{path}/{uuid.uuid4().hex}'


@pytest.fixture
def child2(path):
    yield f'{path}/{uuid.uuid4().hex}'


@pytest.fixture
async def child_watcher(zk, path, child1, child2):
    await zk.create(path)
    watcher = zk.recipes.ChildrenWatcher()
    watcher.set_client(zk)
    yield watcher

    watcher.cancel()

    try:
        await zk.delete(child1)
        await zk.delete(child2)
    except NoNode:
        pass
    await zk.delete(path)


@pytest.mark.asyncio
async def test_child_watch(child_watcher, path, zk, child1, child2):
    children = set()
    ready = asyncio.Event()

    async def children_callback(c):
        for child in c:
            children.add(child)
            ready.set()

    child_watcher.add_callback(path, children_callback)
    assert children == set()
    await zk.create(child1)
    await asyncio.wait([ready.wait()], timeout=0.1)
    assert children == {child1.split('/')[-1]}
    ready.clear()
    await zk.create(child2)
    await asyncio.wait([ready.wait()], timeout=0.1)
    assert ready.is_set()
    assert children == {child.split('/')[-1] for child in (child1, child2)}


@pytest.mark.asyncio
async def test_child_watch_no_node(child_watcher, path):
    random_path = path + uuid.uuid4().hex
    is_finished = asyncio.Future()

    async def stub_callback(d):
        assert d == NoNode
        is_finished.set_result(True)

    child_watcher.add_callback(random_path, stub_callback)
    await asyncio.wait_for(is_finished, 0.1)



import os
import uuid

import pytest

from aiozk import ZKClient, exc  # noqa


HOST = os.environ.get('ZK_HOST', 'zk')


@pytest.fixture
def event_loop(event_loop):
    event_loop.set_debug(True)
    yield event_loop


def get_client():
    return ZKClient(HOST, chroot='/test_aiozk')


async def get_tree(client, curr='/'):
    out = [curr, ]
    children = await client.get_children(curr)
    for c in children:
        # eliminate double slash: //root = '/'.join('/', 'root')
        if curr == '/':
            curr = ''
        out.extend(await get_tree(client, '/'.join([curr, c])))
    return out


async def dump_tree(client, base='/'):
    out = list(await get_tree(client, base))
    print(f'Tree dump: {out}')
    return out


@pytest.fixture
def path():
    yield f'/{uuid.uuid4().hex}'


@pytest.fixture
async def zk():
    c = get_client()
    await c.start()
    if len(await c.get_children('/')):
        await c.deleteall('')
        await c.create('')
    yield c
    try:
        await c.delete('/')
    except exc.NotEmpty:
        await dump_tree(c)
        await c.deleteall('')
        raise
    await c.close()


@pytest.fixture
async def zk2(zk):
    c = get_client()
    await c.start()
    yield c
    await c.close()



from .part import Part
from .primitives import Long, Int


class Stat(Part):
    """
    """
    parts = (
        ("created_zxid", Long),
        ("last_modified_zxid", Long),
        ("created", Long),
        ("modified", Long),
        ("version", Int),
        ("child_version", Int),
        ("acl_version", Int),
        ("ephemeral_owner", Long),
        ("data_length", Int),
        ("num_children", Int),
        ("last_modified_children", Long),
    )


class StatPersisted(Part):
    """
    """
    parts = (
        ("created_zxid", Long),
        ("last_modified_zxid", Long),
        ("created", Long),
        ("modified", Long),
        ("version", Int),
        ("child_version", Int),
        ("acl_version", Int),
        ("ephemeral_owner", Long),
        ("last_modified_children", Long),
    )



# from tornado import gen

from aiozk import exc


class Recipe(object):

    sub_recipes = {}

    def __init__(self, base_path="/"):
        self.client = None
        self.base_path = base_path

        for attribute_name, recipe_class in self.sub_recipes.items():
            recipe_args = ["base_path"]
            if isinstance(recipe_class, tuple):
                recipe_class, recipe_args = recipe_class

            recipe_args = [getattr(self, arg) for arg in recipe_args]

            recipe = recipe_class(*recipe_args)

            setattr(self, attribute_name, recipe)

    def set_client(self, client):
        self.client = client
        for sub_recipe in self.sub_recipes:
            getattr(self, sub_recipe).client = client

    @classmethod
    def validate_dependencies(cls):
        return True

    async def ensure_path(self):
        await self.client.ensure_path(self.base_path)

    async def create_znode(self, path):
        try:
            await self.client.create(path)
        except exc.NodeExists:
            pass
        except exc.NoNode:
            try:
                await self.ensure_path()
            except exc.NodeExists:
                pass



from .request import Request
from .response import Response
from .stat import Stat
from .primitives import UString, Bool


class ExistsRequest(Request):
    """
    """
    opcode = 3

    parts = (
        ("path", UString),
        ("watch", Bool),
    )


class ExistsResponse(Response):
    """
    """
    opcode = 3

    parts = (
        ("stat", Stat),
    )



def drain(iterable):
    """
    Helper method that empties an iterable as it is iterated over.

    Works for:

    * ``dict``
    * ``collections.deque``
    * ``list``
    * ``set``
    """
    if getattr(iterable, "popleft", False):
        def next_item(coll):
            return coll.popleft()
    elif getattr(iterable, "popitem", False):
        def next_item(coll):
            return coll.popitem()
    else:
        def next_item(coll):
            return coll.pop()

    while True:
        try:
            yield next_item(iterable)
        except (IndexError, KeyError):
            return



from .request import Request
from .response import Response
from .stat import Stat
from .primitives import Long, UString


class ReconfigRequest(Request):
    """
    """
    opcode = 16

    parts = (
        ("joining_servers", UString),
        ("leaving_servers", UString),
        ("new_members", UString),
        ("current_config_id", Long),
    )


class ReconfigResponse(Response):
    """
    """
    opcode = 16

    parts = (
        ("stat", Stat),
    )



import asyncio
import logging
import time

from aiozk import exc

from .sequential import SequentialRecipe
from functools import partial

log = logging.getLogger(__name__)


class Lease(SequentialRecipe):

    def __init__(self, base_path, limit=1):
        super(Lease, self).__init__(base_path)
        self.limit = limit

    async def obtain(self, duration):
        lessees = await self.client.get_children(self.base_path)

        if len(lessees) >= self.limit:
            return False

        time_limit = time.time() + duration.total_seconds()

        try:
            await self.create_unique_znode("lease", data=str(time_limit))
        except exc.NodeExists:
            log.warning("Lease for %s already obtained.", self.base_path)

        callback = partial(asyncio.ensure_future, self.release(), loop=self.client.loop)
        self.client.loop.call_later(time_limit, callback)
        return True

    async def release(self):
        await self.delete_unique_znode("lease")



import struct


class Primitive(object):
    """
    The most basic structure of the protocol.  Subclassed, never used directly.

    Used as a building block for the various actually-used primitives outlined
    in the Zookeeper jute file:

    https://github.com/apache/zookeeper/blob/trunk/src/zookeeper.jute
    """
    fmt = None

    def __init__(self, value):
        self.value = value

    def render(self):
        """
        Returns a two-element tuple with the ``struct`` format and list value.

        The value is wrapped in a list, as there are some primitives that deal
        with multiple values.  Any caller of `render()` should expect a list.
        """
        return self.fmt, [self.value]

    @classmethod
    def parse(cls, buff, offset):
        """
        Given a buffer and offset, returns the parsed value and new offset.

        Uses the ``format`` class attribute to unpack the data from the buffer
        and determine the used up number of bytes.
        """
        primitive_struct = struct.Struct("!" + cls.fmt)

        value = primitive_struct.unpack_from(buff, offset)[0]
        offset += primitive_struct.size

        return value, offset

    def __eq__(self, other):
        """
        Basic equality method that tests equality of the ``value`` attributes.
        """
        return self.value == other.value

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.value)


class VariablePrimitive(Primitive):
    """
    Base primitive for variable-length scalar primitives (strings and bytes).
    """
    size_primitive = None

    def render_value(self, value):
        raise NotImplementedError

    @classmethod
    def parse_value(cls, value):
        raise NotImplementedError

    def render(self):
        """
        Returns the ``struct`` format and list of the size and value.

        The format is derived from the size primitive and the length of the
        resulting encoded value (e.g. the format for a string of 'foo' ends
        up as 'h3s'.

        .. note ::
          The value is expected to be string-able (wrapped in ``str()``) and is
          then encoded as UTF-8.
        """
        size_format = self.size_primitive.fmt

        if self.value is None:
            return size_format, [-1]

        value = self.render_value(self.value)

        size = len(value)

        fmt = "%s%ds" % (size_format, size)

        return fmt, [size, value]

    @classmethod
    def parse(cls, buff, offset):
        """
        Given a buffer and offset, returns the parsed value and new offset.

        Parses the ``size_primitive`` first to determine how many more bytes to
        consume to extract the value.
        """
        size, offset = cls.size_primitive.parse(buff, offset)
        if size == -1:
            return None, offset

        var_struct = struct.Struct("!%ds" % size)

        value = var_struct.unpack_from(buff, offset)[0]
        value = cls.parse_value(value)
        offset += var_struct.size

        return value, offset


class Bool(Primitive):
    """
    Represents a boolean (true or false) value.

    Renders as an unsigned char (1 byte).
    """
    fmt = "?"


class Byte(Primitive):
    """
    Represents a single 8-bit byte.
    """
    fmt = "b"


class Int(Primitive):
    """
    Represents an 32-bit signed integer.
    """
    fmt = "i"


class Long(Primitive):
    """
    Represents an 64-bit signed integer.
    """
    fmt = "q"


class Float(Primitive):
    """
    Represents a single-precision floating poing conforming to IEEE 754.
    """
    fmt = "f"


class Double(Primitive):
    """
    Represents a double-precision floating poing conforming to IEEE 754.
    """
    fmt = "d"


class UString(VariablePrimitive):
    """
    Represents a unicode string value, length denoted by a 32-bit integer.
    """
    size_primitive = Int

    def render_value(self, value):
        return bytes(str(value).encode("utf-8"))

    @classmethod
    def parse_value(cls, value):
        return value.decode("utf-8")

    def __str__(self):
        return str(self.value)


class Buffer(VariablePrimitive):
    """
    Represents a bytestring value, length denoted by a 32-bit signed integer.
    """
    size_primitive = Int

    def render_value(self, value):
        if isinstance(value, str):
            return value.encode()
        # return bytes(value)
        return value

    @classmethod
    def parse_value(cls, value):
        return value


class Vector(Primitive):
    """
    Represents an array of any arbitrary `Primitive` or ``Part``.

    Not used directly but rather by its ``of()`` classmethod to denote an
    ``Vector.of(<something>)``.
    """
    item_class = None

    @classmethod
    def of(cls, part_class):
        """
        Creates a new class with the ``item_class`` attribute properly set.
        """
        copy = type(
            "VectorOf%s" % part_class.__name__,
            cls.__bases__, dict(cls.__dict__)
        )
        copy.item_class = part_class

        return copy

    def render(self):
        """
        Creates a composite ``struct`` format and the data to render with it.

        The format and data are prefixed with a 32-bit integer denoting the
        number of elements, after which each of the items in the array value
        are ``render()``-ed and added to the format and data as well.
        """
        value = self.value
        if value is None:
            value = []

        fmt = [Int.fmt]
        data = [len(value)]

        for item_value in value:
            if issubclass(self.item_class, Primitive):
                item = self.item_class(item_value)
            else:
                item = item_value

            item_format, item_data = item.render()
            fmt.extend(item_format)
            data.extend(item_data)

        return "".join(fmt), data

    @classmethod
    def parse(cls, buff, offset):
        """
        Parses a raw buffer at offset and returns the resulting array value.

        Starts off by `parse()`-ing the 32-bit element count, followed by
        parsing items out of the buffer "count" times.
        """
        count, offset = Int.parse(buff, offset)

        values = []
        for _ in range(count):
            value, new_offset = cls.item_class.parse(buff, offset)

            values.append(value)
            offset = new_offset

        return values, offset

    def __str__(self):
        return "%s[%s]" % (
            self.item_class.__name__, ", ".join(map(str, self.value))
        )



import logging
from importlib import import_module

from .recipe import Recipe


log = logging.getLogger(__name__)


RECIPES = {
    'data_watcher': 'DataWatcher',
    'children_watcher': 'ChildrenWatcher',
    'lock': 'Lock',
    'shared_lock': 'SharedLock',
    'lease': 'Lease',
    'barrier': 'Barrier',
    'double_barrier': 'DoubleBarrier',
    'election': 'LeaderElection',
    'party': 'Party',
    'counter': 'Counter',
    'tree_cache': 'TreeCache',
    'allocator': 'Allocator',
}

class RecipeClassProxy(object):

    def __init__(self, client, recipe_class):
        self.client = client
        self.recipe_class = recipe_class

    def __call__(self, *args, **kwargs):
        recipe = self.recipe_class(*args, **kwargs)
        recipe.set_client(self.client)
        return recipe


class RecipeProxy(object):

    def __init__(self, client):
        self.client = client

        self.installed_classes = {}
        self.gather_installed_classes()

    def __getattr__(self, name):
        if name not in self.installed_classes:
            raise AttributeError("No such recipe: %s" % name)

        return RecipeClassProxy(self.client, self.installed_classes[name])

    def gather_installed_classes(self):
        for module, name in RECIPES.items():
            recipe_class = getattr(import_module('aiozk.recipes.{}'.format(module)), name)

            if not issubclass(recipe_class, Recipe):
                log.error("Could not load recipe %s: not a Recipe subclass", recipe_class.__name__)
                continue

            if not recipe_class.validate_dependencies():
                log.error("Could not load recipe %s has unmet dependencies", recipe_class.__name__)
                continue

            log.debug("Loaded recipe %s", recipe_class.__name__)
            self.installed_classes[recipe_class.__name__] = recipe_class



from setuptools import setup, find_packages

from aiozk import __version__


setup(
    name='aiozk',
    version=__version__,
    description='Asyncio client for Zookeeper.',
    author='Kirill Pinchuk',
    author_email='cybergrind@gmail.com',
    maintainer='Kirill Pinchuk',
    maintainer_email='cybergrind@gmail.com',
    url='http://github.com/tipsi/aiozk',
    license='MIT',
    keywords=['zookeeper', 'asyncio', 'async'],
    packages=find_packages(exclude=['tests', 'tests.*', 'docker', 'TAGS', '.projectile', '*.log']),
    install_requires=[
        'tipsi_tools>=0.9.0',
    ],
    entry_points={
    },
    tests_require=[
        'coverage',
        'flake8',
        'pytest',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
    ],
)
